import gc
import time
import board
from digitalio import DigitalInOut, Direction, Pull
import soundstuff


class PinKey:
    def __init__(self, pinid, name):
        self.pin = None
        self.val = False
        self.pinid = pinid
        self.downval = False
        self.name = name
        self.initpin()

    def initpin(self):
        self.pin = DigitalInOut(self.pinid)
        self.pin.direction = Direction.INPUT
        self.pin.pull = Pull.UP

    def deinitpin(self):
        self.pin.deinit()
        self.pin = None

    def isdown(self):
        self.val = self.pin.value == self.downval
        return self.val


class BrailleKeyboard:
    def __init__(self, globaldata):
        gc.collect()
        self.globaldata = globaldata
        self.pins: list[PinKey] = []
        self.sixkeynames = "B1B2B3B4B5B6"
        self.dotkeynames = "B1B2B3B4B5B6B7B8"
        self.modkeynames = "BLB9B0BR"
        self.modkeysused = False
        self.modmodsused = False
        self.down_all = ""
        self.last_all = ""
        self.down_dot = ""
        self.last_dot = ""
        self.down_mod = ""
        self.last_mod = ""
        self.prev_mod = ""
        self.last_key_time = 0
        self.autorepeated = False
        self.threshold1 = self.globaldata.keyrepeatthreshold1
        self.threshold2 = self.globaldata.keyrepeatthreshold2
        self.keydelay1 = self.globaldata.keydelay1
        self.keydelay2 = self.globaldata.keydelay2
        self.delay_time = 0
        self.autopress_time = 0
        self.fastcount = 0
        self.fastcount_beep = 10
        self.allkeys: list[str] = []
        self.lastallkeys: list[str] = []
        self.anykeydown = False
        self.anydotdown = False
        self.anysixdown = False
        self.anymoddown = False
        self.dotdowntime = 0
        self.moddowntime = 0
        self.keyspressed = ""
        self.dotspressed = ""
        self.sixspressed = ""
        self.modifierstr = ""
        self.modspressed = ""
        self.pinschanged = False
        self.dotschanged = False
        self.modschanged = False
        self.checkcount = 0
        self.changecount = 0
        self.last_key = ""
        self.keyqueue = []

        self.B1 = PinKey(board.A4, "B1")
        self.B2 = PinKey(board.A5, "B2")
        self.B3 = PinKey(board.MISO, "B3")
        self.B7 = PinKey(board.SCK, "B7")
        self.B4 = PinKey(board.A0, "B4")
        self.B5 = PinKey(board.A1, "B5")
        self.B6 = PinKey(board.A3, "B6")
        self.B8 = PinKey(board.A2, "B8")
        self.BL = PinKey(board.D12, "BL")
        self.B9 = PinKey(board.D9, "B9")
        self.B0 = PinKey(board.D7, "B0")
        self.BR = PinKey(board.D13, "BR")
        self.addpin(self.BL)
        self.addpin(self.BR)
        self.addpin(self.B9)
        self.addpin(self.B0)
        self.addpin(self.B1)
        self.addpin(self.B2)
        self.addpin(self.B3)
        self.addpin(self.B7)
        self.addpin(self.B4)
        self.addpin(self.B5)
        self.addpin(self.B6)
        self.addpin(self.B8)

    def addpin(self, pin):
        self.pins.append(pin)

    def initpins(self):
        for p in self.pins:
            p.initpin()

    def deinitpins(self):
        for p in self.pins:
            p.deinitpin()

    def addtoallkeys(self, key):
        if key in self.allkeys:
            return
        self.allkeys.append(key)
        # print("addtoallkeys:", self.allkeys)

    def allkey_str(self):
        sortkeys = []
        for k in self.allkeys:
            sortkeys.append(k)
        sortkeys.sort()
        s = ""
        for k in sortkeys:
            s += k
        return s

    def alldot_str(self):
        sortkeys = []
        for k in self.allkeys:
            if k in self.dotkeynames:
                sortkeys.append(k)
        sortkeys.sort()
        s = ""
        for k in sortkeys:
            s += k
        return s

    def allsix_str(self):
        sortkeys = []
        for k in self.allkeys:
            if k in self.sixkeynames:
                sortkeys.append(k)
        sortkeys.sort()
        s = ""
        for k in sortkeys:
            s += k
        return s

    def allmod_str(self, caller):
        modkeys = []
        for k in self.allkeys:
            if k in self.modkeynames:
                modkeys.append(k)
        s = ""
        if "BL" in modkeys:
            s += "BL"
        if "B9" in modkeys:
            s += "B9"
        if "B0" in modkeys:
            s += "B0"
        if "BR" in modkeys:
            s += "BR"
        return s

    def deldotkeys(self):
        newkeys = []
        for k in self.allkeys:
            if k not in self.dotkeynames:
                newkeys.append(k)
        self.allkeys = newkeys

    def checkpins(self):
        self.keyspressed = ""
        self.dotspressed = ""
        self.sixspressed = ""
        self.modspressed = ""
        self.down_all = ""
        self.down_dot = ""
        self.down_mod = ""
        self.anykeydown = False
        for pin in self.pins:
            if pin.isdown():
                self.anykeydown = True
                self.down_all += pin.name
        if self.last_all == self.down_all:
            self.pinschanged = False
            self.dotschanged = False
            self.modschanged = False
            return
        self.anydotdown = False
        self.anysixdown = False
        self.anymoddown = False
        self.last_all = self.down_all
        self.last_key_time = time.monotonic()
        self.pinschanged = True
        self.changecount += 1
        if self.globaldata.keyclick:
            print("new last_str:", self.last_all, self.globaldata.keyclicktone[0], self.globaldata.keyclicktone[1])
            soundstuff.beep(self.globaldata.keyclicktone[0], self.globaldata.keyclicktone[1])
        ns = time.monotonic()
        down_dot = ""
        down_mod = ""
        self.lastallkeys = self.allkeys
        for pin in self.pins:
            if pin.val:
                self.addtoallkeys(pin.name)
                if pin.name in self.modkeynames:
                    self.anymoddown = True
                    down_mod += pin.name
                if pin.name in self.dotkeynames:
                    self.anydotdown = True
                    down_dot += pin.name
                if pin.name in self.sixkeynames:
                    self.anysixdown = True
        if self.last_dot != down_dot:
            self.last_dot = down_dot
            self.dotschanged = True
        if self.last_mod != down_mod:
            self.prev_mod = self.last_mod
            self.last_mod = down_mod
            self.modschanged = True
        if self.anymoddown:
            if self.moddowntime == 0:
                self.moddowntime = ns
        if self.anydotdown:
            if self.dotdowntime == 0:
                self.dotdowntime = ns
        else:
            self.modifierstr = self.allmod_str("1")
            self.dotspressed = self.alldot_str()
        if not self.anysixdown:
            self.modifierstr = self.allmod_str("2")
            self.sixspressed = self.allsix_str()
        if not self.anymoddown:
            self.modspressed = self.allmod_str("3")
        if not self.anykeydown:
            self.modifierstr = self.allmod_str("4")
            self.keyspressed = self.allkey_str()
            self.lastallkeys = self.allkeys
            self.allkeys = []

    def putkey(self, caller, key):
        if self.autorepeated:
            self.autorepeated = False
            print(caller + " putkey:", key, " autorepeated!")
            return
        if len(key) == 0:
            print(caller + " putkey: len(ley)==0 !")
            return
        self.keyqueue.append(key)
        print(caller + " putkey:", key, "len_keyqueue:", len(self.keyqueue))

    def getkey(self):
        self.checkkeys()
        if len(self.keyqueue) == 0:
            return None
        else:
            self.last_key = self.keyqueue.pop(0)
            # print("   getkey:", self.last_key)
            return self.last_key

    def checkkeys(self):
        self.checkcount += 1
        self.checkpins()
        if self.pinschanged:
            if False:
                print("allkeys    :", bool(self.allkeys), self.allkeys)
                print("keyspressed:", bool(self.keyspressed), self.keyspressed)
                print("dotspressed:", bool(self.dotspressed), self.dotspressed)
                print("modspressed:", bool(self.modspressed), self.modspressed)
                print("modifierstr:", bool(self.modifierstr), self.modifierstr)
                print(30*"-", self.changecount)
            if self.dotspressed:
                self.deldotkeys()
                if self.modifierstr:
                    print("modkeysused = True")
                    self.modkeysused = True
                # print(self.dotdowntime, self.moddowntime)
                twokeys = False
                if self.moddowntime == 0:
                    print("dots only ------------------------------------------------")
                else:
                    while True:
                        if self.globaldata.maxmoddiff == 0:
                            print("always 1 key", self.globaldata.maxmoddiff, "--------------------------------")
                            break
                        if self.dotdowntime == self.moddowntime:
                            if self.globaldata.maxmoddiff < 0:
                                print("same time 2 key", self.globaldata.maxmoddiff, "--------------------------------")
                                twokeys = True
                            else:
                                print("same time 1 key", self.globaldata.maxmoddiff, "--------------------------------")
                            break
                        if self.globaldata.maxmoddiff < 0:
                            if self.moddowntime < self.dotdowntime:
                                timediff = abs(self.moddowntime-self.dotdowntime)
                                if timediff < abs(self.globaldata.maxmoddiff):
                                    print("mods earlier 2 key", timediff, self.globaldata.maxmoddiff, "--------------------------------")
                                    twokeys = True
                                else:
                                    print("mods earlier 1 key", timediff, self.globaldata.maxmoddiff, "--------------------------------")
                            else:
                                print("mods later 1 key", timediff, self.globaldata.maxmoddiff, "--------------------------------")
                                twokeys = True
                            break
                        if self.globaldata.maxmoddiff > 0:
                            if self.dotdowntime < self.moddowntime:
                                timediff = abs(self.moddowntime-self.dotdowntime)
                                if timediff > self.globaldata.maxmoddiff:
                                    print("dots earlier 2 key", timediff, self.globaldata.maxmoddiff, "--------------------------------")
                                    twokeys = True
                                else:
                                    print("dots earlier 1 key", timediff, self.globaldata.maxmoddiff, "--------------------------------")
                            else:
                                print("mods earlier 1 key", 0, self.globaldata.maxmoddiff, "--------------------------------")
                            break
                        print("oops !! nothing earlier !!")
                        break
                if twokeys:  # here first the dots and then later the thumb key
                    self.putkey("1a", self.dotspressed)
                    self.putkey("1b", self.modifierstr)
                else:
                    self.putkey("1", self.modifierstr + self.dotspressed)
                self.dotdowntime = 0
            if self.modspressed:
                # print("self.moddowntime = 0:", self.modspressed)
                self.moddowntime = 0
                if self.modkeysused or self.modmodsused or self.anydotdown:
                    pass
                    print("NO KEY self.modkeysused or self.modmodsused or self.anydotdown")
                else:
                    self.putkey("2b", self.modspressed)
                self.modkeysused = False
                self.modmodsused = False
                return
            if self.modschanged:
                print("modschanged currently:", self.last_mod, self.prev_mod, self.lastallkeys)
        else:
            # Autorepeat vvvvvv
            if self.last_all != "":
                if self.anymoddown and not self.anydotdown:
                    if self.modkeysused:
                        return
                    if self.modmodsused:
                        return
                time_diff = time.monotonic() - self.last_key_time
                repeat_time = 0
                if time_diff > self.threshold2:
                    repeat_time = self.keydelay2
                elif time_diff > self.threshold1:
                    repeat_time = self.keydelay1
                    self.fastcount = 0
                if repeat_time > 0:
                    next_time = time.monotonic() - self.autopress_time
                    if next_time > repeat_time:
                        self.autopress_time = time.monotonic()
                        self.autorepeated = False
                        ok = True
                        if self.last_all == "B0B9":
                            ok = False
                        if self.last_all == "B9B0":
                            ok = False
                        if ok:
                            self.putkey("4", self.last_all)
                            self.autorepeated = True
                            if repeat_time == self.keydelay2:
                                self.fastcount += 1
                                if self.fastcount >= self.fastcount_beep:
                                    self.fastcount = 0
                                    try:
                                        soundstuff.beep(self.globaldata.fastrepeattone[0], self.globaldata.fastrepeattone[1])
                                    except:
                                        soundstuff.beep(220, 0.06)


X3 = PinKey(board.SCL,  "X1")
X4 = PinKey(board.SDA,  "X2")
X5 = PinKey(board.TX,   "X3")
X6 = PinKey(board.RX,   "X4")
X8 = PinKey(board.MOSI, "X6")
# XL = [X3, X4, X5, X6, X8]
